/* 
 * xpixmap.c --
 *
 *	This file includes the quotation from Tkwin source file
 * tkbitmap.c.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Software Research Associates not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  Software Research
 * Associates makes no representations about the suitability of this software
 * for any purpose.  It is provided "as is" without express or implied
 * warranty.
 */ 

#include "twinint.h"
#include <x11/xutil.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#define BMP_PLANES 1
#define BMP_PIXELS 1

static HBITMAP BitmapForWindows(char *, unsigned *, unsigned *, int *, int *);
static char *ReadBitmapFile (char *, unsigned *, unsigned *, int *, int *);
static BYTE rebirth(BYTE);

/*
 * XCopyArea, XCopyPlane --
 *
 */

void  
FAR PASCAL
XCopyArea(Display *dummy1, Drawable src, Drawable dest, GC dummy2, 
	int src_x, int src_y, unsigned int width, unsigned int height, 
	int dest_x, int dest_y)
{
    HDC hsrc, hdest;

    hdest = GetDrawableDC(dest);
    hsrc = GetDrawableDC(src);

    BitBlt( hdest, dest_x, dest_y, width, height, hsrc,  src_x,  src_y, SRCCOPY );

    ReleaseDrawableDC(src, hsrc);
    ReleaseDrawableDC(dest, hdest);
}

void FAR PASCAL
XCopyPlane(Display *dummy1, Drawable src, Drawable dest, GC gc,
	int src_x, int src_y, unsigned int width, unsigned int height, 
	int dest_x, int dest_y, unsigned long dummy2)
{
    HDC hsrc, hdest;
    HBRUSH hbr;

    hdest = GetDrawableDC(dest);
    hsrc = GetDrawableDC(src);
    hbr = SelectObject(hdest, CreateSolidBrush(gc->foreground));

    BitBlt( hdest, dest_x, dest_y, width, height, hsrc,  src_x,  src_y,  0xB8074A);

    DeleteObject(SelectObject(hdest, hbr));
    ReleaseDrawableDC(src, hsrc);
    ReleaseDrawableDC(dest, hdest);
}

/*
 * XCreateBitmapFromData, XReadBitmapFile --
 *
 */

Pixmap
FAR PASCAL
XCreateBitmapFromData(Display *dummy1, Drawable dummy2, const char *data,
	unsigned int width, unsigned int height)
{
    int nByte;
    int v, h, adjust;
    BYTE *bPtr;
    HBITMAP hbmp;

    nByte = (width + 7) / 8;
    adjust = nByte % 2;

    bPtr = (BYTE *) malloc((nByte + adjust) * height);
    for (v = 0; (unsigned) v <	height; v++) {
        for (h = 0; h < nByte; h++) {
	    bPtr[v * (nByte + adjust) + h] = 
	    	    rebirth((BYTE)data[v * nByte + h]); 
        }
	if (adjust) bPtr[v * (nByte + adjust) + h] = '\0';
    }
    hbmp = CreateBitmap(width, height, BMP_PLANES, BMP_PIXELS, bPtr);
    free(bPtr);

    return hbmp ;
}

static BYTE 
rebirth(BYTE byte)
{
    int i;
    BYTE work = (BYTE) ~byte, result = 0, bit = 0x80;

    for (i = 0; i < 8; i++) {
   	if (work & 0x01) {
   	    result |= bit;
   	}
   	work >>= 1;
   	bit >>= 1;
   }
   return result;
}     	

int FAR PASCAL 
XReadBitmapFile(Display *dummy1, Drawable dummy2, const char *filename,
	unsigned int *width, unsigned int *height, Pixmap *bitmap,
	int *x_hot, int *y_hot)
{
    struct _stat st;
    char *buffer;
    FILE *fp;
    HBITMAP hbmp;
    BITMAPFILEHEADER *bmfh;
    int result;

    if ((fp = fopen(filename, "rb")) == NULL)	   
 	return BitmapOpenFailed;
    if (_fstat(fileno(fp), &st) == -1) {				   
 	result = BitmapOpenFailed;				   
	goto close_end;
    }
    if ((buffer = (char *)malloc((size_t)st.st_size + 1)) == NULL) {
 	result = BitmapNoMemory;
	goto close_end;
    }
    if (fread(buffer, (size_t)st.st_size, 1, fp) != 1) {	
 	result = BitmapFileInvalid;
	goto  free_end;
    }
    buffer[st.st_size] = '\0';

    bmfh = (BITMAPFILEHEADER *)buffer;
    if (strncmp((char *)&bmfh->bfType, "BM", 2) == 0) {	 
 	hbmp = BitmapForWindows(buffer, width, height, x_hot, y_hot);
    } else {
	char *data;
	data = ReadBitmapFile((char *)filename, width, height, x_hot, y_hot);
	if (data == NULL) {
	    goto free_end;
 	    result = BitmapFileInvalid;
	}
	hbmp = CreateBitmap(*width, *height, 1, 1, data);
	free(data);
    }

    if (hbmp == NULL) {
 	result =  BitmapFileInvalid;
	goto free_end;
    }
    *bitmap = hbmp;
    result = BitmapSuccess;
free_end:
    free(buffer);	 
close_end:
    fclose(fp);
    return result;
}	

static HBITMAP
BitmapForWindows(char *data, unsigned *width, unsigned *height, int *x_hot, int *y_hot)
{
    BITMAPFILEHEADER *bmfh = (BITMAPFILEHEADER *) data;
    BITMAPINFOHEADER *bmih = (BITMAPINFOHEADER *)(data + sizeof(BITMAPFILEHEADER));
    BITMAPINFO *bmi = (BITMAPINFO *)bmih;
    char *bits = data + bmfh->bfOffBits;		
    HBITMAP hbmp;
    HDC hdc;

    hdc = GetDC(NULL);
    hbmp = CreateDIBitmap(hdc, bmih, CBM_INIT, bits, bmi, DIB_RGB_COLORS);	 
    ReleaseDC(NULL, hdc);
    *width = (int) bmih->biWidth;
    *height = (int) bmih->biHeight;
    if (!x_hot) *x_hot = -1;
    if (!y_hot) *y_hot = -1;
    return hbmp;
}

/*
 * XCreatePixmap, XFreePixmap --
 *
 */

Pixmap FAR PASCAL
XCreatePixmap(Display *dummy1, Drawable dummy2, 
	unsigned int width, unsigned int height, unsigned int dummy3)
{
#if 1
    HBITMAP hbmp;
    HDC hdc;

    hdc = GetDC(NULL);
    hbmp = CreateCompatibleBitmap(hdc, width, height);
    ReleaseDC(NULL, hdc);

    return hbmp;

#else
    BITMAPINFOHEADER bmih;
    HBITMAP hBitmap;
    HDC hdc;
    size_t size = sizeof(BITMAPINFOHEADER);
  
    memset(&bmih, 0, size);
    bmih.biSize = (DWORD) size;
    bmih.biWidth = (DWORD) width;
    bmih.biHeight = (DWORD) height;
    bmih.biPlanes = 1;
    bmih.biBitCount = 8;
    hdc = GetDC(NULL);
    hBitmap = CreateDIBitmap(hdc, &bmih, 0L, NULL, NULL, 0);
    ReleaseDC(NULL, hdc);
    return hBitmap;
#endif
}

void FAR PASCAL
XFreePixmap(Display *dummy, Pixmap pixmap)
{
    if (pixmap != NULL) DeleteObject(pixmap);
}

/* 
 * twBitBlt --
 *
 */

#undef BitBlt
#if 0
BOOL 
FAR PASCAL
twBitBlt(HDC hdest, int dest_x, int dest_y, int width, int height, 
	HDC hsrc,  int src_x,  int src_y, DWORD rop)
{
    int v, h, d, w;
    BOOL result;
    long lim = (16000L * 8) / nColorres;
    
    d = (lim > (long) width) ? width : (int) lim;
    for (v = 0; v < height; v++) {
	for (h = 0; h < width; h += d) {
	    w = (width > (h + d)) ? d : width - h;
	    result = BitBlt( hdest, h, v, w, 1, hsrc,  src_x + h,  src_y + v, rop);
	    if (result == 0) return result;
	}
    }
    return result;
}
#else
BOOL 
FAR PASCAL
twBitBlt(HDC hdest, int dest_x, int dest_y, int width, int height, 
	HDC hsrc,  int src_x,  int src_y, DWORD rop)
{
    int cw, ch, uw, uh, tw, th;
    BOOL result;
    long lim = (16000L * 8) / nColorres;
    
    uh = (int) (lim / (long) width);
    uh = uh ? uh : 1;
    uw = (int) (lim < width ? lim : width);

    for (ch = 0; ch < height; ch += uh) {
    	tw = ((ch + uh) < height) ? uh : (height - ch);
	for (cw = 0; cw < width; cw += uw) {
	    tw =  ((cw + uw) < width) ? uw : (width - cw);
	    result = BitBlt(hdest, dest_x + cw, dest_y + ch, tw, th, hsrc,  
	    	    src_x + cw, src_y + ch, rop);
	    if (result == 0) return result;
	}
    }
    return result;
}
#endif
#define TCL_MSDOS_PORT
#define sscanf twsscanf
/*
 * The code below is quoted from the Tkwin source file tkbitmap.c.
 *					-- haga@sra.co.jp
 */


#ifdef TCL_MSDOS_PORT
/* The code below is slightly modified from the X11R5 Xlib source file
 * XRdBitF.c
 */
/* $XFree86: mit/lib/X/XRdBitF.c,v 1.4 1993/03/27 09:14:21 dawes Exp $ */
/* $XConsortium: XRdBitF.c,v 1.15 91/02/01 16:34:46 gildea Exp $ */
/* Copyright, 1987, Massachusetts Institute of Technology */

/*
Permission to use, copy, modify, distribute, and sell this software and its
documentation for any purpose is hereby granted without fee, provided that
the above copyright notice appear in all copies and that both that
copyright notice and this permission notice appear in supporting
documentation, and that the name of M.I.T. not be used in advertising or
publicity pertaining to distribution of the software without specific,
written prior permission.  M.I.T. makes no representations about the
suitability of this software for any purpose.  It is provided "as is"
without express or implied warranty.
*/

#define MAX_SIZE 255

/* shared data for the image read/parse logic */
static short hexTable[256];		/* conversion value */
static Bool Xinitialized = False;	/* easier to fill in at run time */


/*
 *	Table index for the hex values. Initialized once, first time.
 *	Used for translation value or delimiter significance lookup.
 */
static void initHexTable()
{
    hexTable['0'] = 0;	hexTable['1'] = 1;
    hexTable['2'] = 2;	hexTable['3'] = 3;
    hexTable['4'] = 4;	hexTable['5'] = 5;
    hexTable['6'] = 6;	hexTable['7'] = 7;
    hexTable['8'] = 8;	hexTable['9'] = 9;
    hexTable['A'] = 10;	hexTable['B'] = 11;
    hexTable['C'] = 12;	hexTable['D'] = 13;
    hexTable['E'] = 14;	hexTable['F'] = 15;
    hexTable['a'] = 10;	hexTable['b'] = 11;
    hexTable['c'] = 12;	hexTable['d'] = 13;
    hexTable['e'] = 14;	hexTable['f'] = 15;

    /* delimiters of significance are flagged w/ negative value */
    hexTable[' '] = -1;	hexTable[','] = -1;
    hexTable['}'] = -1;	hexTable['\n'] = -1;
    hexTable['\t'] = -1;
	
    Xinitialized = True;
}

/*
 *	read next hex value in the input stream, return -1 if EOF
 */
static long NextInt (fstream)
    FILE *fstream;
{
    int	ch;
    long value = 0;
    int gotone = 0;
    int done = 0;
    
    /* loop, accumulate hex value until find delimiter  */
    /* skip any initial delimiters found in read stream */

    while (!done) {
	ch = getc(fstream);
	if (ch == EOF) {
	    value	= -1L;
	    done++;
	} else {
	    /* trim high bits, check type and accumulate */
	    ch &= 0xff;
	    if (isascii(ch) && isxdigit(ch)) {
		value = (value << 4) + hexTable[ch];
		gotone++;
	    } else if ((hexTable[ch]) < 0 && gotone)
	      done++;
	}
    }
    return value;
}


/* invert the order of the bits in a character, and complement */
static swap(int c){
    int i, r = 0;
    for(i=0; i < 8; i++){
	r <<= 1;
	if(c&1) r |= 1;
	c >>= 1;
    }
    return (~r)&0xff;
}

/* $XConsortium: Xlib.h,v 11.215 91/07/22 15:42:38 rws Exp $ */
/* 
 * Copyright 1985, 1986, 1987, 1991 by the Massachusetts Institute of Technology
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided 
 * that the above copyright notice appear in all copies and that both that 
 * copyright notice and this permission notice appear in supporting 
 * documentation, and that the name of M.I.T. not be used in advertising
 * or publicity pertaining to distribution of the software without specific, 
 * written prior permission. M.I.T. makes no representations about the 
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * X Window System is a Trademark of MIT.
 *
 */

/* read bitmap from X10 or X11 bitmap files and return
 * array of data bits in format appropriate for Microsoft Windows
 */
static char *ReadBitmapFile (filename, width, height, x_hot, y_hot)
    char *filename;
    unsigned int *width, *height;       /* RETURNED */
    int *x_hot, *y_hot;                 /* RETURNED */
{
    FILE *fstream;			/* handle on file  */
    char *data = NULL;			/* working variable */
    char line[MAX_SIZE];		/* input line from file */
#ifdef TCL_MSDOS_PORT
    unsigned size;			/* number of bytes of data */
#else
    int size;				/* number of bytes of data */
#endif
    char name_and_type[MAX_SIZE];	/* an input line */
    char *type;				/* for parsing */
    long value;				/* from an input line */
    int val;
    int version10p;			/* boolean, old format */
    int padding;			/* to handle alignment */
    int bytes_per_line;			/* per scanline of data */
    unsigned int ww = 0;		/* width */
    unsigned int hh = 0;		/* height */
    int hx = -1;			/* x hotspot */
    int hy = -1;			/* y hotspot */

    /* first time initialization */
    if (Xinitialized == False) initHexTable();

    if ((fstream = fopen(filename, "rt")) == NULL) {
	return NULL;
    }

    /* error cleanup and return macro	*/
#define	RETURN(code) { if (data) free (data); fclose (fstream); return code; }

    while (fgets(line, MAX_SIZE, fstream)) {
	if (strlen(line) == MAX_SIZE-1) {
	    RETURN (NULL);
	}
	if (sscanf(line,"#define %s %d",name_and_type,&val) == 2) {
	    if ((type = strrchr(name_and_type, '_')) == NULL)
	      type = name_and_type;
	    else
	      type++;

	    if (!strcmp("width", type))
	      ww = (unsigned int) val;
	    if (!strcmp("height", type))
	      hh = (unsigned int) val;
	    if (!strcmp("hot", type)) {
		if (type-- == name_and_type || type-- == name_and_type)
		  continue;
		if (!strcmp("x_hot", type))
		  hx = val;
		if (!strcmp("y_hot", type))
		  hy = val;
	    }
	    continue;
	}
    
	if (sscanf(line, "static short %s = {", name_and_type) == 1)
	  version10p = 1;
	else if (sscanf(line,"static unsigned char %s = {",name_and_type) == 1)
	  version10p = 0;
	else if (sscanf(line, "static char %s = {", name_and_type) == 1)
	  version10p = 0;
	else
	  continue;

	if ((type = strrchr(name_and_type, '_')) == NULL)
	  type = name_and_type;
	else
	  type++;

	if (strcmp("bits[]", type))
	  continue;
    
	if (!ww || !hh)
	  RETURN (NULL);

	if ((ww % 16) && ((ww % 16) < 9) && !version10p)
	  padding = 1;
	else padding = 0;

	bytes_per_line = (ww+7)/8 + padding;

	size = bytes_per_line * hh;
/* xxxxxxxxxx added +hh into size; but then removed it */
	data = (char *) malloc ((unsigned int) (size));
	if (!data) 
	  RETURN (NULL);

	if (version10p) {
	    unsigned char *ptr;
#ifdef _WINDOWS
	    unsigned bytes;
#else
	    int bytes;
#endif
	    for (bytes=0, ptr=data; bytes<size; (bytes += 2)) {
		if ((value = NextInt(fstream)) < 0)
		  RETURN (NULL);
#ifdef _WINDOWS
		*(ptr++) = (unsigned char) swap((int) (value & 0xff));
		*(ptr++) = (unsigned char) swap((int) ((value >> 8)&0xff));
#else
		*(ptr++) = swap(value & 0xff);
		*(ptr++) = swap((value >> 8)&0xff);
#endif
	    }
	} else {
	    unsigned char *ptr;
#ifdef _WINDOWS
	    unsigned bytes;
#else
	    int bytes;
#endif

	    for (bytes=0, ptr=data; bytes<size; bytes++, ptr++) {
		if ((value = NextInt(fstream)) < 0) 
		  RETURN (NULL);
#ifdef _WINDOWS
		*ptr=(unsigned char) swap((int) (value&0xff));
		if(padding && ((bytes+2)%bytes_per_line) == 0){
		    *ptr++ = (unsigned char) '\377';
#else
		*ptr=swap(value&0xff);
		if(padding && ((bytes+2)%bytes_per_line) == 0){
		    *ptr++ = '\377';
#endif
		    bytes++;
		}
	    }
	}
    }					/* end while */


    *width = ww;
    *height = hh;
    if (x_hot) *x_hot = hx;
    if (y_hot) *y_hot = hy;

    RETURN (data);
}
#endif
